@@ -7,12 +7,14 @@ use File::Spec::Functions; # catfile
 use MIME::Base64;
 use File::Path qw/make_path/;
 use LWP::UserAgent;
+use HTTP::Request;
+use HTTP::Headers;
 use JSON::MaybeXS;
 use YAML;
 use URI::Encode qw/uri_encode/;
 use Encode;
 
-our $VERSION = 0.015;
+our $VERSION = 0.016;
 
 =head1 NAME
 
@@ -114,6 +116,14 @@ has '_zendesk_credentials' => (
     lazy	=> 1,
     builder	=> '_build_zendesk_credentials',
     );
+    
+has 'default_headers' => (
+    is		=> 'ro',
+    isa		=> 'HTTP::Headers',
+    required	=> 1,
+    lazy	=> 1,
+    builder	=> '_build_default_headers',
+    );
 
 sub _build_user_agent {
     my $self = shift;
@@ -121,12 +131,19 @@ sub _build_user_agent {
     my $ua = LWP::UserAgent->new(
 	keep_alive	=> 1
     );
-    $ua->default_header( 'Content-Type'	    => "application/json" );
-    $ua->default_header( 'Accept'	    => "application/json" );
-    $ua->default_header( 'Authorization'    => "Basic " . $self->_zendesk_credentials );
+   # $ua->default_headers( $self->default_headers );
     return $ua;
 }
 
+sub _build_default_headers {
+    my $self = shift;
+    my $h = HTTP::Headers->new();
+    $h->header( 'Content-Type'	=> "application/json" );
+    $h->header( 'Accept'	=> "application/json" );
+    $h->header( 'Authorization' => "Basic " . $self->_zendesk_credentials );
+    return $h;
+}
+
 sub _build_zendesk_credentials {
     my $self = shift;
     return encode_base64( $self->zendesk_username . "/token:" . $self->zendesk_token );
@@ -311,16 +328,16 @@ sub download_attachment {
 	}
     }
     
-    my $response = $self->user_agent->get( 
-	$params{attachment}{content_url},
-	':content_file'	=> $target, 
-	# So we don't get a http 406 error
-	'Content-Type'	=> '',
-	'Accept'	=> '',
-	);
-    if( not $response->is_success ){
-	$self->log->logdie( "Zendesk API Error: http status:".  $response->code .' '.  $response->message );
-    }
+    # Empty headers so we don't get a http 406 error
+    my $headers = $self->default_headers->clone();
+    $headers->header( 'Content-Type'  => '' );
+    $headers->header( 'Accept'        => '' );
+    $self->_request_from_api(
+        method      => 'GET',
+        uri         => $params{attachment}{content_url},
+        headers     => $headers,
+        fields  => { ':content_file' => $target },
+        );
     return $target;
 }
 
@@ -398,12 +415,13 @@ sub update_ticket {
     );
 
     my $encoded_body = encode_json( $params{body} );
-    $self->log->trace( "Submitting:\n" . $encoded_body );
+    $self->log->trace( "Submitting:\n" . $encoded_body ) if $self->log->is_trace;
+
     my $response = $self->_request_from_api(
-            method  => 'put',
-	    path    => '/tickets/' . $params{ticket_id} . '.json',
-	    body    => $encoded_body,
-	);
+        method          => 'PUT',
+        path            => '/tickets/' . $params{ticket_id} . '.json',
+        body            => $encoded_body,
+        );
     $self->cache_set( 'ticket-' . $params{ticket_id}, $response->{ticket} ) unless( $params{no_cache} );
     return $response;
 }
@@ -440,9 +458,9 @@ sub get_ticket {
     if( not $ticket ){
 	$self->log->debug( "Ticket info not cached, requesting fresh: $params{ticket_id}" );
 	my $info = $self->_request_from_api(
-            method  => 'get',
-	    path    => '/tickets/' . $params{ticket_id} . '.json',
-	);
+            method      => 'GET',
+            path        => '/tickets/' . $params{ticket_id} . '.json',
+            );
 	
 	if( not $info or not $info->{ticket} ){
 	    $self->log->logdie( "Could not get ticket info for ticket: $params{ticket_id}" );
@@ -551,9 +569,9 @@ sub get_organization {
     if( not $organization ){
 	$self->log->debug( "Organization info not in cache, requesting fresh: $params{organization_id}" );
 	my $info = $self->_request_from_api(
-            method  => 'get',
-	    path    => '/organizations/' . $params{organization_id} . '.json',
-	);
+            method      => 'GET',
+            path        => '/organizations/' . $params{organization_id} . '.json',
+            );
 	if( not $info or not $info->{organization} ){
 	    $self->log->logdie( "Could not get organization info for organization: $params{organization_id}" );
 	}
@@ -668,12 +686,12 @@ sub update_organization {
     };
 
     my $encoded_body = encode_json( $body );
-    $self->log->trace( "Submitting:\n" . $encoded_body );
+    $self->log->trace( "Submitting:\n" . $encoded_body ) if $self->log->is_trace;
     my $response = $self->_request_from_api(
-        method  => 'put',
-	    path    => '/organizations/' . $params{organization_id} . '.json',
-	    body    => $encoded_body,
-	);
+        method      => 'PUT',
+        path        => '/organizations/' . $params{organization_id} . '.json',
+        body        => $encoded_body,
+        );
     if( not $response or not $response->{organization}{id} == $params{organization_id} ){
 	$self->log->logdie( "Could not update organization: $params{organization_id}" );
     }
@@ -845,12 +863,13 @@ sub update_user {
     };
 
     my $encoded_body = encode_json( $body );
-    $self->log->trace( "Submitting:\n" . $encoded_body );
+    $self->log->trace( "Submitting:\n" . $encoded_body ) if $self->log->is_trace;
     my $response = $self->_request_from_api(
-        method  => 'put',
-            path    => '/users/' . $params{user_id} . '.json',
-            body    => $encoded_body,
+        method      => 'PUT',
+        path        => '/users/' . $params{user_id} . '.json',
+        body        => $encoded_body,
         );
+
     if( not $response or not $response->{user}{id} == $params{user_id} ){
         $self->log->logdie( "Could not update user: $params{user_id}" );
     }
@@ -959,10 +978,9 @@ sub _paged_get_request_from_api {
     my $response = undef;
     do{
         $response = $self->_request_from_api(
-            method  => 'get',
-	    path    => $params{path} . ( $params{path} =~ m/\?/ ? '&' : '?' ) . 'page=' . $page,
-	);
-	
+            method      => 'GET',
+            path        => $params{path} . ( $params{path} =~ m/\?/ ? '&' : '?' ) . 'page=' . $page,
+            );
 	push( @results, @{ $response->{$params{field} } } );
 	$page++;
       }while( $response->{next_page} and ( not $params{size} or scalar( @results ) < $params{size} ) );
@@ -975,28 +993,54 @@ sub _request_from_api {
     my ( $self, %params ) = validated_hash(
         \@_,
         method	=> { isa => 'Str' },
-	path	=> { isa => 'Str' },
+	path	=> { isa => 'Str', optional => 1 },
+        uri     => { isa => 'Str', optional => 1 },
         body    => { isa => 'Str', optional => 1 },
+        headers => { isa => 'HTTP::Headers', optional => 1 },
+        fields  => { isa => 'HashRef', optional => 1 },
+        
     );
-    my $url = $self->zendesk_api_url . $params{path};
-    $self->log->debug( "Requesting from Zendesk ($params{method}): $url" );
+    my $url;
+    if( $params{uri} ){
+        $url = $params{uri};
+    }elsif( $params{path} ){
+        $url =  $self->zendesk_api_url . $params{path};
+    }else{
+        $self->log->logdie( "Cannot request without either a path or uri" );
+    }
+
+    my $request = HTTP::Request->new(
+        $params{method},
+        $url,
+        $params{headers} || $self->default_headers,
+        );
+    $request->content( $params{body} ) if( $params{body} );
+
+    $self->log->debug( "Requesting from Zendesk: " . $request->uri );
+    $self->log->trace( "Request:\n" . Dump( $request ) ) if $self->log->is_trace;
+
     my $response;
-    
     my $retry = 1;
     my $retry_delay = $self->default_backoff;
     do{
-        if( $params{method} =~ m/^get$/i ){
-            $response = $self->user_agent->get( $url );
-
-        }elsif( $params{method} =~ m/^put$/i ){
-            if( $params{body} ){
-                $response = $self->user_agent->put( $url,
-                                                    Content => $params{body} );
-            }else{
-                $response = $self->user_agent->put( $url );
+        # Fields are a special use-case for GET requests:
+        # https://metacpan.org/pod/LWP::UserAgent#ua-get-url-field_name-value
+        if( $params{fields} ){
+            if( $request->method ne 'GET' ){
+                $self->log->logdie( 'Cannot use fields unless the request method is GET' );
             }
+            my %fields = %{ $params{fields} };
+            my $headers = $request->headers();
+            foreach( keys( %{ $headers } ) ){
+                $fields{$_} = $headers->{$_};
+            }
+            $self->log->trace( "Fields:\n" . Dump( \%fields ) );
+            $response = $self->user_agent->get(
+                $request->uri(),
+                %fields,
+            );
         }else{
-            $self->log->logdie( "Unsupported request method: $params{method}" );
+            $response = $self->user_agent->request( $request );
         }
         if( not $response->is_success ){
             if(  $response->code == 503 ){
@@ -1020,6 +1064,8 @@ sub _request_from_api {
                 $self->log->warn( "Received a 500 (Server closed connection without sending any data)... going to backoff and retry!");
             }elsif( $response->code == 500 and $response->decoded_content =~ m/read timeout/ ){
                 $self->log->warn( "Received a 500 (read timeout)... going to backoff and retry!");
+            }elsif( $response->code == 504 and $response->decoded_content =~ m/Gateway Time-out/ ){
+                $self->log->warn( "Received a 504 (Gateway Time-out)... going to backoff and retry!");
             }else{
                 $retry = 0;
             }
@@ -1029,12 +1075,15 @@ sub _request_from_api {
             }
         }
     }while( $retry and not $response );
+
+    $self->log->trace( "Zendesk Response:\n", Dump( $response ) ) if $self->log->is_trace;
     if( not $response->is_success ){
 	$self->log->logdie( "Zendesk API Error: http status:".  $response->code .' '.  $response->message . ' Content: ' . $response->content);
     }
-    $self->log->trace( "Zendesk API Error: http status:".  $response->code .' '.  $response->message . ' Content: ' . $response->content);
-
-    return decode_json( encode( 'utf8', $response->decoded_content ) );
+    if( $response->decoded_content ){
+        return decode_json( encode( 'utf8', $response->decoded_content ) );
+    }
+    return;
 }
 
 
